@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 # BSD 3-Clause License
 #
-# Copyright (c) 2020-2022, Faster Speeding
+# Copyright (c) 2020-2023, Faster Speeding
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -42,7 +42,6 @@ import subprocess  # noqa: S404
 import sys
 import tempfile
 import time
-import tomllib
 import traceback
 import types
 import typing
@@ -55,14 +54,14 @@ import dotenv
 import fastapi
 import httpx
 import jwt
-import pydantic
+import piped_shared
 import starlette.middleware
 from anyio.streams import memory as streams
 from asgiref import typing as asgiref
 from dateutil import parser as dateutil
 
 if sys.version_info < (3, 11):
-    raise RuntimeError("Only supports Python 3.10+")
+    raise RuntimeError("Only supports Python 3.11+")
 
 _LOGGER = logging.getLogger("piped.bot")
 _LOGGER.setLevel("INFO")
@@ -86,12 +85,6 @@ WEBHOOK_SECRET = os.environ["WEBHOOK_SECRET"]
 jwt_instance = jwt.JWT()
 
 
-class _Config(pydantic.BaseModel):
-    bot_actions: set[str] = pydantic.Field(
-        default_factory=lambda: {"Freeze PR dependency changes", "Resync piped", "Reformat PR code"}
-    )
-
-
 class _ProcessingIndex:
     """Index of the PRs being processed."""
 
@@ -640,11 +633,6 @@ async def post_webhook(
     return fastapi.Response(status_code=204)
 
 
-def _read_toml(path: pathlib.Path) -> dict[str, typing.Any]:
-    with path.open("rb") as file:
-        return tomllib.load(file)
-
-
 @contextlib.asynccontextmanager
 async def _with_cloned(
     output: typing.IO[str], url: str, /, *, branch: str = "master"
@@ -824,8 +812,7 @@ async def _process_repo(
         run_ctx = _RunCheck(http, token=token, repo_name=full_name, commit_hash=head_sha)
 
         async with run_ctx, _with_cloned(run_ctx.output, git_url, branch=head_ref) as temp_dir_path:
-            pyproject = await anyio.to_thread.run_sync(_read_toml, temp_dir_path / "pyproject.toml")
-            config = _Config.parse_obj(pyproject["tool"]["piped"])
+            config = await piped_shared.Config.read_async(temp_dir_path)
             if not config.bot_actions:
                 _LOGGER.warn("Received event from %s repo with no bot_wait_for", full_name)
                 return
@@ -894,12 +881,13 @@ async def _apply_patch(
         # If this conflicted then we should allow another CI run to redo these
         # changes after the current changes have been pushed.
         except subprocess.CalledProcessError:
-            pass
+            await anyio.to_thread.run_sync(pathlib.Path(patch_path).unlink)
 
         else:
+            await anyio.to_thread.run_sync(pathlib.Path(patch_path).unlink)
+            await run_process(output, ["git", "add", "."], cwd=cwd, env=COMMIT_ENV)
             await run_process(output, ["git", "commit", "-am", workflow.name], cwd=cwd, env=COMMIT_ENV)
 
-        await anyio.to_thread.run_sync(pathlib.Path(patch_path).unlink)
         break
 
 
@@ -933,9 +921,3 @@ async def run_process(
 
     else:
         output.write(result.stdout.decode())
-
-
-# TODO: for some reason this is getting stuck on a background task while trying to stop it
-# TODO: this probably shouldn't run on a PR with conflicts.
-# TODO: edge cases where the token expires mid-run.
-# TODO: rerun requests.
