@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 # BSD 3-Clause License
 #
-# Copyright (c) 2020-2022, Faster Speeding
+# Copyright (c) 2020-2023, Faster Speeding
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -50,110 +50,25 @@ __all__: typing.List[str] = [
     "verify_types",
 ]
 
+import datetime
 import itertools
 import pathlib
 import re
-import shutil
 import typing
 from collections import abc as collections
 
 import nox
-import pydantic
+import piped_shared
 import tomli
 
 _CallbackT = typing.TypeVar("_CallbackT", bound=collections.Callable[..., typing.Any])
-ConfigEntryT = typing.Union[dict[str, str], list[str], str]
-ConfigT = dict[str, ConfigEntryT]
-
-
-def _default_version():
-    requires: typing.Optional[str]
-    if (project := _pyproject_toml.get("project")) and (requires := project.get("requires-python")):
-        return requires.lstrip(">=")
-
-    return "3.9,<3.12"
-
-
-class _Config(pydantic.BaseModel):
-    """Configuration class for the project config."""
-
-    codespell_ignore: typing.Optional[str] = None
-    default_sessions: list[str]
-    dep_locks: list[pathlib.Path] = pydantic.Field(default_factory=lambda: [pathlib.Path("./dev-requirements/")])
-    extra_test_installs: list[str] = pydantic.Field(default_factory=list)
-    github_actions: typing.Union[dict[str, ConfigT], list[str]] = pydantic.Field(
-        default_factory=lambda: ["resync-piped"]
-    )
-    hide: list[str] = pydantic.Field(default_factory=list)
-    mypy_allowed_to_fail: bool = False
-    mypy_targets: list[str] = pydantic.Field(default_factory=list)
-
-    # Right now pydantic fails to recognise Pattern[str] so we have to hide this
-    # at runtime.
-    if typing.TYPE_CHECKING:
-        path_ignore: typing.Optional[re.Pattern[str]] = None
-
-    else:
-        path_ignore: typing.Optional[re.Pattern] = None
-
-    project_name: typing.Optional[str] = None
-    top_level_targets: list[str]
-    version_constraint: str = pydantic.Field(default_factory=_default_version)
-
-    def assert_project_name(self) -> str:
-        if not self.project_name:
-            raise RuntimeError("This CI cannot run without project_name")
-
-        return self.project_name
-
-    def codespell_ignore_args(self) -> list[str]:
-        if self.codespell_ignore:
-            return ["--ignore-regex", self.codespell_ignore]
-
-        return []
-
-
-with pathlib.Path("pyproject.toml").open("rb") as _file:
-    _pyproject_toml = tomli.load(_file)
-    _config = _Config.parse_obj(_pyproject_toml["tool"]["piped"])
-
 
+_config = piped_shared.Config.read(pathlib.Path("./"))
 nox.options.sessions = _config.default_sessions
 _DEPS_DIR = pathlib.Path("./dev-requirements")
 _SELF_INSTALL_REGEX = re.compile(r"^\.\[.+\]$")
 
 
-def _exists(path: pathlib.Path) -> typing.Optional[pathlib.Path]:
-    return path if path.exists() else None
-
-
-def _dev_path(value: str, /) -> pathlib.Path:
-    return _exists(_DEPS_DIR / f"{value}.txt") or pathlib.Path(__file__).parent / "base-requirements" / f"{value}.txt"
-
-
-def _constraints_txt() -> typing.Optional[pathlib.Path]:
-    return _exists(_DEPS_DIR / "constraints.txt")
-
-
-def _runtime_deps() -> list[str]:
-    path = _constraints_txt()
-    if path:
-        actual_file = _exists(_DEPS_DIR / "constraints.txt")
-        if not actual_file:
-            raise RuntimeError("Couldn't find constraints.txt")
-
-        return ["-r", str(actual_file), "-c", str(path)]
-
-    return []
-
-
-def _deps(*dev_deps: str, constrain: bool = False) -> collections.Iterator[str]:
-    if constrain and (path := _constraints_txt()):
-        return itertools.chain(["-c", str(path)], _deps(*dev_deps))
-
-    return itertools.chain.from_iterable(("-r", str(_dev_path(value))) for value in dev_deps)
-
-
 def _tracked_files(session: nox.Session, *, force_all: bool = False) -> collections.Iterable[str]:
     output = session.run("git", "--no-pager", "grep", "--threads=1", "-l", "", external=True, log=False, silent=True)
     assert isinstance(output, str)
@@ -164,21 +79,65 @@ def _tracked_files(session: nox.Session, *, force_all: bool = False) -> collecti
     return output.splitlines()
 
 
-def _install_deps(session: nox.Session, *requirements: str, first_call: bool = True) -> None:
-    if not requirements:
+def _dev_path(value: str, /) -> typing.Optional[pathlib.Path]:
+    for path in [_DEPS_DIR / f"{value}.txt", pathlib.Path(__file__).parent / "base-requirements" / f"{value}.txt"]:
+        if path.exists():
+            return path
+
+    return None
+
+
+def _other_dep(name: str) -> str:
+    return f"other-{name}"
+
+
+def _install_deps(
+    session: nox.Session, *requirements_tuple: str, names: list[str] = [], first_call: bool = True
+) -> None:
+    if not requirements_tuple and not names:
         return
 
+    requirements = list(requirements_tuple)
+    other_requirements: dict[pathlib.Path, list[str]] = {}
+    for name in names:
+        if name != "constraints":
+            path = _dev_path(name)
+            assert path, "If this doesn't exist then path is missing the base requirement and is broken"
+            requirements.extend(["-r", str(path)])
+            if path := _dev_path(_other_dep(name)):
+                files = other_requirements.get(path.parent) or []
+                files.extend(["-r", path.name])
+                other_requirements[path.parent] = files
+
+            continue
+
+        path = _DEPS_DIR / "constraints.txt"
+        if path.exists():
+            requirements.extend(["-r", str(path)])
+
+        other_path = path.with_name(_other_dep(path.name))
+        if other_path.exists():
+            files = other_requirements.get(other_path.parent) or []
+            files.extend(["-r", other_path.name])
+            other_requirements[other_path.parent] = files
+
     # --no-install --no-venv leads to it trying to install in the global venv
     # as --no-install only skips "reused" venvs and global is not considered reused.
     if not _try_find_option(session, "--skip-install", when_empty="True"):
         if first_call:
             session.install("--upgrade", "wheel")
 
-        session.install("--upgrade", *map(str, requirements))
+        session.install("--upgrade", *requirements)
 
     elif any(map(_SELF_INSTALL_REGEX.fullmatch, requirements)):
         session.install("--upgrade", "--force-reinstall", "--no-dependencies", ".")
 
+    for cwd, paths in other_requirements.items():
+        # We override the cwd using --root to handle relative dep links as relative to the requirement file.
+        # TODO: do we want to keep no-deps here?
+        with session.chdir(cwd):  # TODO: get "--root", str(cwd) to work
+            session.install("--upgrade", "--no-deps", *paths)
+
 
 def _try_find_option(
     session: nox.Session, name: str, /, *other_names: str, when_empty: typing.Optional[str] = None
@@ -227,7 +186,7 @@ def cleanup(session: nox.Session) -> None:
     import shutil
 
     # Remove directories
-    raw_paths = ["./dist", "./site", "./.nox", "./.pytest_cache", "./coverage_html"]
+    raw_paths = ["./dist", "./site", "./.nox", "./.pytest_cache", "./coverage_html", ".mypy_cache"]
     if _config.project_name:
         raw_paths.append(f"{_config.project_name}.egg-info")
 
@@ -255,8 +214,15 @@ def cleanup(session: nox.Session) -> None:
             session.log(f"[  OK  ] Removed '{raw_path}'")
 
 
-_ACTION_DEFAULTS = {"DEFAULT_PY_VER": "3.9", "NOX_DEP_PATH": "./piped/python/base-requirements/nox.txt"}
-_resync_filter: list[str] = ["piped"]
+_DEFAULT_COMMITER_USERNAME = "always-on-duty[bot]"
+
+_ACTION_DEFAULTS = {
+    "ACTION_COMMITTER_EMAIL": f"120557446+{_DEFAULT_COMMITER_USERNAME}@users.noreply.github.com",
+    "ACTION_COMMITTER_USERNAME": _DEFAULT_COMMITER_USERNAME,
+    "DEFAULT_PY_VER": "3.9",
+    "NOX_DEP_PATH": "./piped/python/base-requirements/nox.txt",
+}
+_resync_filter: list[str] = ["piped", "pyproject.toml", "requirements.in"]
 _verify_filter: list[str] = []
 _dep_locks: list[pathlib.Path] = []
 
@@ -283,18 +249,21 @@ if _config.dep_locks:
 class _Action:
     __slots__ = ("defaults", "required_names")
 
-    def __init__(self, *, required: collections.Sequence[str] = (), defaults: typing.Optional[ConfigT] = None) -> None:
-        self.defaults: ConfigT = dict(_ACTION_DEFAULTS)
+    def __init__(
+        self, *, required: collections.Sequence[str] = (), defaults: typing.Optional[piped_shared.ConfigT] = None
+    ) -> None:
+        self.defaults: piped_shared.ConfigT = dict(_ACTION_DEFAULTS)
         self.defaults.update(defaults or ())
         self.required_names = frozenset(required or ())
 
-    def process_config(self, config: ConfigT, /) -> ConfigT:
-        output: ConfigT = dict(self.defaults)
+    def process_config(self, config: piped_shared.ConfigT, /) -> piped_shared.ConfigT:
+        output: piped_shared.ConfigT = dict(self.defaults)
         output.update(**config)
         return output
 
 
 _ACTIONS: dict[str, _Action] = {
+    "clippy": _Action(),
     "freeze-for-pr": _Action(defaults={"EXTEND_FILTERS": [], "FILTERS": _resync_filter}),
     "lint": _Action(),
     "pr-docs": _Action(),
@@ -304,9 +273,11 @@ _ACTIONS: dict[str, _Action] = {
         defaults={"CODECLIMATE_TOKEN": "", "OSES": "[ubuntu-latest, macos-latest, windows-latest]"},
     ),
     "reformat": _Action(),
-    "release-docs": _Action(),
+    "release-docs": _Action(defaults={"BRANCH_PUSHES": None}),
     "resync-piped": _Action(defaults={"FILTERS": ["piped", "pyproject.toml"]}),
+    "rustfmt": _Action(),
     "type-check": _Action(),
+    "update-licence": _Action(),
     "upgrade-locks": _Action(),
     "verify-locks": _Action(defaults={"EXTEND_FILTERS": [], "FILTERS": _verify_filter}),
     "verify-types": _Action(),
@@ -328,10 +299,14 @@ def copy_actions(_: nox.Session) -> None:
     to_write: dict[pathlib.Path, str] = {}
     if isinstance(_config.github_actions, dict):
         actions = iter(_config.github_actions.items())
-        wild_card: collections.ItemsView[str, ConfigEntryT] = (_config.github_actions.get("*") or {}).items()
+        wild_card: collections.ItemsView[str, piped_shared.ConfigEntryT] = (
+            _config.github_actions.get("*") or {}
+        ).items()
 
     else:
-        actions: collections.Iterable[tuple[str, ConfigT]] = ((name, {}) for name in _config.github_actions)
+        actions: collections.Iterable[tuple[str, piped_shared.ConfigT]] = (
+            (name, {}) for name in _config.github_actions
+        )
         wild_card = {}.items()
 
     for file_name, config in actions:
@@ -350,14 +325,20 @@ def copy_actions(_: nox.Session) -> None:
         full_config = action.process_config(config)
         to_write[pathlib.Path("./.github/workflows") / file_name] = template.render(**full_config, config=_config)
 
+    pathlib.Path("./.github/workflows").mkdir(exist_ok=True, parents=True)
+
     for path, value in to_write.items():
         with path.open("w+") as file:
             file.write(value)
 
 
-def _to_valid_urls(session: nox.Session) -> typing.Optional[set[pathlib.Path]]:
-    if session.posargs:
-        return set(map(pathlib.Path.resolve, map(pathlib.Path, session.posargs)))
+def _pyproject_toml() -> typing.Optional[dict[str, typing.Any]]:
+    try:
+        with pathlib.Path("pyproject.toml").open("rb") as _file:
+            return tomli.load(_file)
+
+    except FileNotFoundError:
+        return None
 
 
 def _freeze_file(session: nox.Session, path: pathlib.Path, /) -> None:
@@ -369,7 +350,7 @@ def _freeze_file(session: nox.Session, path: pathlib.Path, /) -> None:
             "-o",
             str(target),
             "--min-python-version",
-            _config.version_constraint,
+            _config.version(_pyproject_toml()),
             str(path),
         )
 
@@ -380,29 +361,31 @@ _EXTRAS_FILTER = re.compile(r"\[.+\]")
 @nox.session(name="freeze-locks", reuse_venv=True)
 def freeze_locks(session: nox.Session) -> None:
     """Freeze the dependency locks."""
-    _install_deps(session, *_deps("freeze-locks"))
-    valid_urls = _to_valid_urls(session)
+    _install_deps(session, names=["freeze-locks"])
     constraints_lock = pathlib.Path("./dev-requirements/constraints.txt")
+    project = _pyproject_toml()
 
-    if not valid_urls:
-        project = _pyproject_toml.get("project") or {}
-        deps: list[str] = project.get("dependencies") or []
-        if optional := project.get("optional-dependencies"):
-            deps.extend(itertools.chain(*optional.values()))
+    if not project:
+        raise RuntimeError("Missing pyproject.toml")
 
-        tl_requirements = pathlib.Path("./requirements.in")
-        if tl_requirements.exists():
-            with tl_requirements.open("r") as file:
-                deps.extend(file.read().splitlines())
+    project = project.get("project") or {}
+    deps: list[str] = project.get("dependencies") or []
+    if optional := project.get("optional-dependencies"):
+        deps.extend(itertools.chain(*optional.values()))
 
-        constraints_in = pathlib.Path("./dev-requirements/constraints.in")
-        if deps:
-            with constraints_in.open("w+") as file:
-                file.write("\n".join(deps) + "\n")
+    tl_requirements = pathlib.Path("./requirements.in")
+    if tl_requirements.exists():
+        with tl_requirements.open("r") as file:
+            deps.extend(file.read().splitlines())
 
-        else:
-            constraints_in.unlink(missing_ok=True)
-            constraints_lock.unlink(missing_ok=True)
+    constraints_in = pathlib.Path("./dev-requirements/constraints.in")
+    if deps:
+        with constraints_in.open("w+") as file:
+            file.write("\n".join(deps) + "\n")
+
+    else:
+        constraints_in.unlink(missing_ok=True)
+        constraints_lock.unlink(missing_ok=True)
 
     for lock_path in _config.dep_locks:
         if lock_path.is_file():
@@ -427,23 +410,24 @@ def freeze_locks(session: nox.Session) -> None:
 def verify_deps(session: nox.Session) -> None:
     """Verify the dependency locks by installing them."""
     for path in itertools.chain.from_iterable(path.glob("*.txt") if path.is_dir() else (path,) for path in _dep_locks):
-        session.install("--dry-run", "-r", str(path))
+        # We override the cwd using --root to handle relative dep links as relative to the requirement file.
+        # TODO: do we want to keep no-deps here?
+        with session.chdir(path.parent):  # TODO: get "--root", str(cwd) to work
+            session.install("--dry-run", "-r", str(path.relative_to(path.parent)))
 
 
 @_filtered_session(name="generate-docs", reuse_venv=True)
 def generate_docs(session: nox.Session) -> None:
     """Generate docs for this project using Mkdoc."""
-    _install_deps(session, *_deps("docs"))
+    _install_deps(session, names=["docs"])
     output_directory = _try_find_option(session, "-o", "--output") or "./site"
     session.run("mkdocs", "build", "-d", output_directory)
-    for path in ("./CHANGELOG.md", "./README.md"):
-        shutil.copy(path, pathlib.Path(output_directory) / path)
 
 
 @_filtered_session(reuse_venv=True)
 def flake8(session: nox.Session) -> None:
     """Run this project's modules against the pre-defined flake8 linters."""
-    _install_deps(session, *_deps("flake8"))
+    _install_deps(session, names=["flake8"])
     session.log("Running flake8")
     session.run("pflake8", *_config.top_level_targets, log=False)
 
@@ -453,7 +437,7 @@ def slot_check(session: nox.Session) -> None:
     """Check this project's slotted classes for common mistakes."""
     # TODO: don't require installing .?
     # https://github.com/pypa/pip/issues/10362
-    _install_deps(session, *_runtime_deps(), *_deps("lint"))
+    _install_deps(session, names=["constraints", "lint"])
     _install_deps(session, "--no-deps", *_config.extra_test_installs, first_call=False)
     session.run("slotscheck", "-m", _config.assert_project_name())
 
@@ -461,7 +445,7 @@ def slot_check(session: nox.Session) -> None:
 @_filtered_session(reuse_venv=True, name="spell-check")
 def spell_check(session: nox.Session) -> None:
     """Check this project's text-like files for common spelling mistakes."""
-    _install_deps(session, *_deps("lint"))
+    _install_deps(session, names=["lint"])
     session.log("Running codespell")
     session.run("codespell", *_tracked_files(session), *_config.codespell_ignore_args(), log=False)
 
@@ -469,15 +453,46 @@ def spell_check(session: nox.Session) -> None:
 @_filtered_session(reuse_venv=True)
 def build(session: nox.Session) -> None:
     """Build this project using flit."""
-    _install_deps(session, *_deps("publish"))
+    _install_deps(session, names=["publish"])
     session.log("Starting build")
     session.run("flit", "build")
 
 
+_LICENCE_PATTERN = re.compile(r"(Copyright \(c\) (\d+-?\d*))")
+
+
+def _update_licence(match: re.Match[str]) -> str:
+    licence_str, date_range = match.groups()
+    start = date_range.split("-", 1)[0]
+    current_year = str(datetime.datetime.now().year)
+
+    if start == current_year:
+        return licence_str
+
+    return f"{licence_str.removesuffix(date_range)}{start}-{current_year}"
+
+
+_LICENCE_FILE_PATTERN = re.compile(r".*(rs|py)|LICENSE")
+
+
+@_filtered_session(name="update-licence", venv_backend="none")
+def update_licence(session: nox.Session):
+    """Bump the end year of the project's licence to the current year."""
+    for path in map(pathlib.Path, _tracked_files(session)):
+        if not _LICENCE_FILE_PATTERN.fullmatch(path.name):
+            continue
+
+        data = path.read_text()
+        new_data = _LICENCE_PATTERN.sub(_update_licence, data)
+
+        if new_data != data:
+            path.write_text(new_data)
+
+
 @_filtered_session(name="verify-markup", reuse_venv=True)
 def verify_markup(session: nox.Session):
     """Verify the syntax of the repo's markup files."""
-    _install_deps(session, *_deps("lint"))
+    _install_deps(session, names=["lint"])
     tracked_files = list(_tracked_files(session))
 
     session.log("Running pre_commit_hooks.check_toml")
@@ -503,9 +518,7 @@ def verify_markup(session: nox.Session):
 
 def _publish(session: nox.Session, /, *, env: typing.Optional[dict[str, str]] = None) -> None:
     # https://github.com/pypa/pip/issues/10362
-    _install_deps(session, *_runtime_deps(), *_deps("publish"))
-    # TODO: does this need to install .?
-    _install_deps(session, "--no-deps", ".", first_call=False)
+    _install_deps(session, names=["publish"])
 
     env = env or session.env.copy()
     if target := session.env.get("PUBLISH_TARGET"):
@@ -535,16 +548,18 @@ def test_publish(session: nox.Session) -> None:
 @_filtered_session(reuse_venv=True)
 def reformat(session: nox.Session) -> None:
     """Reformat this project's modules to fit the standard style."""
-    _install_deps(session, *_deps("reformat"))
-    session.run("black", *_config.top_level_targets)
-    session.run("isort", *_config.top_level_targets)
-    session.run("pycln", *_config.top_level_targets)
+    _install_deps(session, names=["reformat"])
+    if _config.top_level_targets:
+        session.run("black", *_config.top_level_targets)
+        session.run("isort", *_config.top_level_targets)
+        session.run("pycln", *_config.top_level_targets)
 
     tracked_files = list(_tracked_files(session, force_all=True))
     py_files = [path for path in tracked_files if re.fullmatch(r".+\.pyi?$", path)]
 
-    session.log("Running sort-all")
-    session.run("sort-all", *py_files, success_codes=[0, 1], log=False)
+    if py_files:
+        session.log("Running sort-all")
+        session.run("sort-all", *py_files, success_codes=[0, 1], log=False)
 
     session.log("Running pre_commit_hooks.end_of_file_fixer")
     session.run("python", "-m", "pre_commit_hooks.end_of_file_fixer", *tracked_files, success_codes=[0, 1], log=False)
@@ -559,7 +574,7 @@ def reformat(session: nox.Session) -> None:
 def test(session: nox.Session) -> None:
     """Run this project's tests using pytest."""
     # https://github.com/pypa/pip/issues/10362
-    _install_deps(session, *_deps("tests"))
+    _install_deps(session, names=["tests"])
     _install_deps(session, *_config.extra_test_installs, first_call=False)
     # TODO: can import-mode be specified in the config.
     session.run("pytest", "-n", "auto", "--import-mode", "importlib")
@@ -570,7 +585,7 @@ def test_coverage(session: nox.Session) -> None:
     """Run this project's tests while recording test coverage."""
     project_name = _config.assert_project_name()
     # https://github.com/pypa/pip/issues/10362
-    _install_deps(session, *_deps("tests"))
+    _install_deps(session, names=["tests"])
     _install_deps(session, *_config.extra_test_installs, first_call=False)
     # TODO: can import-mode be specified in the config.
     # https://github.com/nedbat/coveragepy/issues/1002
@@ -594,7 +609,7 @@ def _run_pyright(session: nox.Session, /, *args: str) -> None:
 @_filtered_session(name="type-check", reuse_venv=True)
 def type_check(session: nox.Session) -> None:
     """Statically analyse and veirfy this project using Pyright."""
-    _install_deps(session, *_deps("type-checking"))
+    _install_deps(session, names=["type-checking"])
     _run_pyright(session)
 
     if _config.mypy_targets:
@@ -611,7 +626,7 @@ def verify_types(session: nox.Session) -> None:
     project_name = _config.assert_project_name()
     # TODO is installing . necessary here?
     # https://github.com/pypa/pip/issues/10362
-    _install_deps(session, *_runtime_deps(), *_deps("type-checking"))
+    _install_deps(session, names=["type-checking"])
     _install_deps(session, *_config.extra_test_installs, first_call=False)
     _run_pyright(session, "--verifytypes", project_name, "--ignoreexternal")
 
@@ -626,7 +641,7 @@ def sync_piped(session: nox.Session) -> None:
 def fetch_piped(session: nox.Session) -> None:
     """Fetch Piped from upstream and resync."""
     session.run("git", "submodule", "update", "--remote", "piped", external=True)
-    _install_deps(session, *_deps("nox"))
+    _install_deps(session, names=["nox"])
     # We call this through nox's CLI like this to ensure that the updated version
     # of these sessions are called.
     session.run("nox", "-s", "copy-piped")
